import sys
import yaml
import requests
import tempfile
import shutil
import logging
import collections
import os
import errno
import hashlib
import tarfile
from time import time
from pkg_resources import Requirement, resource_filename, require
from substance.exceptions import (
    ConfigSyntaxError,
    FileSystemError,
    FileDoesNotExist
)

logger = logging.getLogger(__name__)

_yaml_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG


def _dict_representer(dumper, data):
    return dumper.represent_dict(data.iteritems())


def _dict_constructor(loader, node):
    return collections.OrderedDict(loader.construct_pairs(node))


yaml.add_representer(collections.OrderedDict, _dict_representer)
yaml.add_constructor(_yaml_mapping_tag, _dict_constructor)


def writeToFile(filename, data):
    try:
        with open(filename, "wb") as fh:
            fh.write(data)
    except Exception as err:
        raise FileSystemError("Failed to write to %s : %s" % (filename, err))


def writeYAML(filename, data):
    try:
        with open(filename, "w") as fileh:
            fileh.write(yaml.dump(data, default_flow_style=False))
    except Exception as err:
        raise FileSystemError("Failed to write to %s : %s" % (filename, err))


def readYAML(filename):
    try:
        stream = open(filename, "r")
        contents = yaml.load(stream)
        return contents
    except yaml.YAMLError, exc:
        msg = "Syntax error in file %s"
        if hasattr(exc, 'problem_mark'):
            mark = exc.problem_mark
            msg += " Error position: (%s:%s)" % (mark.line +
                                                 1, mark.column + 1)
        raise ConfigSyntaxError(msg)
    except Exception as err:
        raise FileSystemError(
            "Failed to read configuration file %s : %s" % (filename, err))


def getSupportFile(filename):
    return resource_filename(Requirement.parse('substance'), filename)


def getPackageVersion():
    return require('substance')[0].version


def readSupportFile(filename):
    try:
        keyfile = getSupportFile(filename)
        with open(keyfile, "r") as fileh:
            keydata = fileh.read()
            return keydata
    except IOError as err:
        raise FileDoesNotExist("File not found: %s" % keyfile)
    except Exception as err:
        raise FileSystemError("Failed to read: %s" % keyfile)


def streamDownload(url, destination):
    basename = os.path.basename(destination)

    try:
        with open(destination, 'wb') as fd:
            r = requests.get(url, stream=True, timeout=5)
            r.raise_for_status()
            contentLength = int(r.headers.get('content-length'))

            if contentLength is None:
                fd.write(r.content)
            else:
                progress = 0
                startTime = time()
                chunkSize = contentLength / 100
                for chunk in r.iter_content(chunk_size=chunkSize):
                    if chunk:
                        progress += len(chunk)
                        fd.write(chunk)
                        done = int(50 * progress / contentLength)
                        elapsed = (time() - startTime)
                        speed = humanReadableBytes(progress / elapsed)
                        left = humanReadableBytes(contentLength - progress)
                        sys.stdout.write(
                            "\r%s [%s%s] %s/s %s left" % (basename,  '=' * done, ' ' * (50 - done), speed, left))
                        sys.stdout.flush()
                sys.stdout.write("\n")
            fd.flush()
    except (Exception, KeyboardInterrupt) as err:
        if os.path.exists(destination):
            os.remove(destination)
        raise(err)
    return destination


def humanReadableBytes(num):
    for x in ['bytes', 'KB', 'MB', 'GB']:
        if num < 1024.0:
            return "%3.1f%s" % (num, x)
        num /= 1024.0
    return "%3.1f%s" % (num, 'TB')


def sha1sum(filename):
    sha = hashlib.sha1()
    bufsize = 65536
    with open(filename, 'rb') as fd:
        while True:
            data = fd.read(bufsize)
            if not data:
                break
            sha.update(data)
    return sha.hexdigest()


def makeHttpRequest(url, params=None):
    r = requests.get(url, params=params)
    r.raise_for_status()
    return r.content


def makeXHRRequest(url, params=None):
    r = requests.get(url, params=params)
    r.raise_for_status()
    payload = r.json()
    return payload


def untar(filename, destination=None):
    tar = tarfile.open(filename)
    tar.extractall(destination)
    tar.close()
    return filename


def pathComponents(path):
    folders = []
    path = os.path.normpath(path)
    while 1:
        path, folder = os.path.split(path)

        if folder != "":
            folders.append(folder)
        else:
            if path != "":
                folders.append(path)
            break

    folders.reverse()
    return folders


def mergeDict(a, b, path=None):
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                mergeDict(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


def readDotEnv(filepath, env={}):
    with open(filepath) as f:
        return parseDotEnv(f, env)


def parseDotEnv(dotenvdata, env={}):
    for line in dotenvdata:
        line = line.strip()
        if not line or line.startswith('#') or '=' not in line:
            continue
        k, v = line.split('=', 1)
        v = v.strip("'").strip('"')
        env[k] = v
    return env


def makeSymlink(source, link, force=False):
    try:
        os.symlink(source, link)
    except OSError, e:
        if e.errno == errno.EEXIST and force:
            os.remove(link)
            os.symlink(source, link)
        else:
            raise e


def readSymlink(link):
    return os.readlink(link)


def expandLocalPath(path, basePath=None):
    path = os.path.expanduser(path)
    if not os.path.isabs(path) and basePath:
        path = os.path.normpath(path)
        path = os.path.join(basePath, path)
    return os.path.abspath(os.path.realpath(path))



import os
import time
from collections import OrderedDict
from substance.monads import *
from substance.constants import *
from substance.utils import readDotEnv, writeToFile, makeSymlink
from substance.exceptions import (InvalidEnvError, InvalidOptionError)
from substance.config import (Config)
from substance import Shell
from substance.subenv.constants import (SPECDIR, ENVFILE, CODELINK, CONFFILE)
import jinja2

logger = logging.getLogger(__name__)


class SubenvSpec(object):
    def __init__(self, specPath, basePath, name=None, vars={}, lastApplied=None):
        self.specPath = specPath
        self.basePath = basePath
        self.envPath = None
        self.name = name
        self.envFiles = []
        self.overrides = vars
        self.vars = OrderedDict()
        self.lastApplied = lastApplied
        self.current = False
        self.struct = {'files': [], 'dirs': []}

    def setEnvPath(self, path):
        self.envPath = path
        self.config = Config(os.path.join(self.envPath, CONFFILE))

    @staticmethod
    def fromEnvPath(path):
        if not os.path.isdir(path):
            return Fail(InvalidOptionError("Specified path '%s' does not exist." % path))

        envPath = path
        name = os.path.basename(envPath)

        envVars = Try.attempt(readDotEnv, os.path.join(envPath, ENVFILE))
        if envVars.isFail():
            return envVars
        envVars = envVars.getOK()

        reserved = ['SUBENV_NAME', 'SUBENV_LASTAPPLIED',
                    'SUBENV_ENVPATH', 'SUBENV_SPECPATH', 'SUBENV_BASEPATH']
        vars = envVars.copy()
        for k in vars.keys():
            if k in reserved:
                del vars[k]

        lastApplied = None
        if 'SUBENV_LASTAPPLIED' in envVars:
            lastApplied = envVars['SUBENV_LASTAPPLIED']

        env = SubenvSpec(envVars['SUBENV_SPECPATH'], envVars['SUBENV_BASEPATH'],
                         envVars['SUBENV_NAME'], vars, lastApplied)
        env.setEnvPath(envPath)
        return env

    @staticmethod
    def fromSpecPath(path, vars={}):
        if not os.path.isdir(path):
            return Fail(InvalidEnvError("Specified path '%s' does not exist." % path))

        if os.path.basename(path) == SPECDIR or not os.path.isdir(os.path.join(path, SPECDIR)):
            return Fail(InvalidOptionError("Invalid path specified. Please pass a path to a folder with a %s directory." % SPECDIR))
        specPath = os.path.join(path, SPECDIR)

        name = os.path.basename(path)

        return SubenvSpec(specPath, path, name, vars).scan()

    def getLastAppliedDateTime(self, fmt='%Y-%m-%d %H:%M:%S'):
        if self.lastApplied:
            return time.strftime(fmt, time.localtime(float(self.lastApplied)))
        return None

    def scan(self):
        return self.loadEnvVars(self.overrides) \
            .then(self.loadEnvStruct)

    def applyTo(self, envPath):
        self.setEnvPath(envPath)
        return self.clearEnv() \
            .then(self.applyDirs) \
            .then(self.applyFiles)  \
            .then(self.writeEnv) \
            .then(self.linkCode) \
            .then(self.assertConfig) \
            .then(self.applyScript) \
            .then(lambda: OK(self))

    def linkCode(self):
        return Try.sequence([
            Try.attempt(makeSymlink, self.basePath,
                        os.path.join(self.envPath, CODELINK), True),
            Try.attempt(makeSymlink, self.specPath,
                        os.path.join(self.envPath, SPECDIR), True)
        ])

    def clearEnv(self):
        if os.path.isfile(self.config.configFile):
            return Shell.rmFile(self.config.configFile)
        return OK(None)

    def writeEnv(self):
        dotenv = os.path.join(self.envPath, ENVFILE)
        logger.debug("Writing environment to: %s" % dotenv)
        envVars = OrderedDict(self.vars, **self.overrides)
        envVars.update({
            'SUBENV_NAME': self.name,
            'SUBENV_LASTAPPLIED': time.time(),
            'SUBENV_ENVPATH': self.envPath,
            'SUBENV_SPECPATH': self.specPath,
            'SUBENV_BASEPATH': self.basePath
        })

        env = "\n".join(["%s=\"%s\"" % (k, v) for k, v in envVars.iteritems()])
        return Try.attempt(writeToFile, dotenv, env)

    def assertConfig(self):
        if not os.path.isfile(self.config.configFile):
            return OK({})
        return self.config.loadConfigFile()

    def applyScript(self):
        commands = self.config.get('script', [])
        return Try.sequence(map(self.applyCommand,  commands))

    def applyCommand(self, cmd):
        logger.info("Running environment command: %s" % cmd)
        return Shell.call(cmd, cwd=self.envPath, shell=True)

    def applyDirs(self):
        ops = [Shell.makeDirectory(self.envPath, 0750)]
        for dir in self.struct['dirs']:
            sourceDir = os.path.join(self.specPath, dir)
            destDir = os.path.join(self.envPath, dir)
            mode = os.stat(sourceDir).st_mode & 0777
            logger.debug("Creating directory '%s' mode: %s" % (dir, mode))
            ops.append(Shell.makeDirectory(destDir).bind(
                defer(Shell.chmod, mode=mode)))

        return Try.sequence(ops)

    def applyFiles(self):
        ops = []
        for file in self.struct['files']:
            fname, ext = os.path.splitext(file)
            source = os.path.join(self.specPath, file)
            dest = os.path.join(self.envPath, file)

            if fname == ENVFILE:
                continue
            elif ext == '.jinja':
                logger.debug("Rendering '%s' to %s" % (file, dest))
                dest = os.path.splitext(dest)[0]
                ops.append(self.renderFile(file, dest, self.vars))
            else:
                logger.debug("Copying '%s' to %s" % (file, dest))
                ops.append(Shell.copyFile(os.path.join(
                    self.specPath, file), os.path.join(self.envPath, file)))

        return Try.sequence(ops)

    def getEnvVars(self):
        envVars = self.overrides.copy()
        envVars.update({
            'SUBENV_NAME': self.name,
            'SUBENV_LASTAPPLIED': self.lastApplied,
            'SUBENV_ENVPATH': self.envPath,
            'SUBENV_SPECPATH': self.specPath,
            'SUBENV_BASEPATH': self.basePath
        })
        return envVars

    def renderFile(self, source, dest, vars={}):
        try:
            tplEnv = jinja2.Environment(
                loader=jinja2.FileSystemLoader(self.specPath))
            tpl = tplEnv.get_template(source)

            tplVars = self.getEnvVars()
            tplVars.update({'subenv': self})
            tplVars.update({'SUBENV_VARS': self.getEnvVars()})

            with open(dest, 'wb') as fh:
                fh.write(tpl.render(**tplVars))
            return OK(None)
        except Exception as err:
            return Fail(err)

    def loadEnvStruct(self):
        return self.scanEnvStruct().bind(self.setEnvStruct)

    def scanEnvStruct(self):
        struct = {'dirs': [], 'files': []}
        for root, dirs, files in os.walk(self.specPath):
            relPath = os.path.relpath(
                root, self.specPath).strip('./').strip('/')
            for dir in dirs:
                struct['dirs'].append(os.path.join(relPath, dir))
            for file in files:
                struct['files'].append(os.path.join(relPath, file))
        return OK(struct)

    def setEnvStruct(self, struct):
        self.struct = struct
        return OK(self)

    def setEnvVars(self, e={}):
        self.overrides = e
        return OK(self)

    def loadEnvVars(self, env={}):
        return self.readEnvFiles() \
            .map(lambda e: dict(e, **env)) \
            .bind(self.setEnvVars)

    def readEnvFiles(self):
        specEnvFile = os.path.join(self.specPath, ENVFILE)
        if os.path.isfile(specEnvFile):
            self.envFiles.append(specEnvFile)

        baseEnvFile = os.path.join(self.basePath, ENVFILE)
        if os.path.isfile(baseEnvFile):
            self.envFiles.append(baseEnvFile)

        map(lambda x: logger.info("Loading dotenv file: '%s'" % x), self.envFiles)
        return Try.sequence(map(Try.attemptDeferred(readDotEnv), self.envFiles))  \
            .map(lambda envs: reduce(lambda acc, x: dict(acc, **x), envs, {}))

    def __repr__(self):
        return "SubEnvSpec(%(name)s) spec:%(specPath)s base:%(basePath)s envFile:%(envFiles)s vars:%(vars)s files:%(struct)s" % self.__dict__



from setuptools import setup, find_packages
import platform

with open('README.rst') as f:
    readme = f.read()

execfile('substance/_version.py')

install_requires = ['setuptools>=1.1.3', 'PyYAML', 'tabulate', 'paramiko>=2.3.1', 'netaddr', 'requests', 'tinydb', 'python_hosts==0.3.3', 'jinja2']

if 'Darwin' in platform.system():
    install_requires.append('macfsevents')

setup(name='substance',
      version=__version__,
      author='Turbulent',
      author_email='oss@turbulent.ca',
      url='https://substance.readthedocs.io/',
      license='Apache License 2.0',
      long_description=readme,
      description='Substance - Local dockerized development environment',
      install_requires=install_requires,
      packages=find_packages(),
      package_data={'substance': ['support/*']},
      test_suite='tests',
      zip_safe=False,
      include_package_data=True,
      entry_points={
          'console_scripts': [
              'substance = substance.cli:cli',
              'subenv = substance.subenv.cli:cli'
          ],
      })
